import json
from typing import Any, Text, Dict

from sqlalchemy import Column, Integer, String, ForeignKey, Boolean
from sqlalchemy.orm import relationship

from rasax.community.database.base import Base


class Domain(Base):
    """Stores the domain of the currently deployed Core model."""

    __tablename__ = "domain"

    id = Column(Integer, primary_key=True)
    project_id = Column(String, ForeignKey("project.project_id"))
    store_entities_as_slots = Column(Boolean, default=True)

    actions = relationship("DomainAction", cascade="all, delete-orphan",
                           back_populates="domain")
    intents = relationship("DomainIntent", cascade="all, delete-orphan",
                           back_populates="domain")
    entities = relationship("DomainEntity", cascade="all, delete-orphan",
                            back_populates="domain")
    slots = relationship("DomainSlot", cascade="all, delete-orphan",
                         back_populates="domain")
    templates = relationship("DomainTemplate", cascade="all, delete-orphan",
                             back_populates="domain")

    def as_dict(self) -> Dict[Text, Any]:
        return {
            "config": {
                "store_entities_as_slots": self.store_entities_as_slots
            },
            "actions": [e.action for e in self.actions if not e.is_form],
            "forms": [e.action for e in self.actions if e.is_form],
            "entities": [e.entity for e in self.entities],
            "intents": [{i.intent: {
                "use_entities": i.use_entities}}
                for i in self.intents],
            "slots": {s.slot: {
                "auto_fill": s.auto_fill,
                "initial_value": s.initial_value,
                "type": s.type} for s in self.slots},
            "templates": {t.template: json.loads(t.content) for t in
                          self.templates}
        }


class DomainAction(Base):
    """Stores the actions which are defined in the domain."""

    __tablename__ = "domain_action"

    id = Column(Integer, ForeignKey("domain.id"), primary_key=True)
    action = Column(String, primary_key=True)
    is_form = Column(Boolean, default=False)

    domain = relationship("Domain", back_populates="actions")


class DomainIntent(Base):
    """Stores the intents which are defined in the domain."""

    __tablename__ = "domain_intent"

    id = Column(Integer, ForeignKey("domain.id"), primary_key=True)
    intent = Column(String, primary_key=True)
    use_entities = Column(Boolean, default=True)

    domain = relationship("Domain", back_populates="intents")


class DomainEntity(Base):
    """Stores the entities which are defined in the domain."""

    __tablename__ = "domain_entity"

    id = Column(Integer, ForeignKey("domain.id"), primary_key=True)
    entity = Column(String, primary_key=True)

    domain = relationship("Domain", back_populates="entities")


class DomainSlot(Base):
    """Stores the slots which are defined in the domain."""

    __tablename__ = "domain_slot"

    id = Column(Integer, ForeignKey("domain.id"), primary_key=True)
    slot = Column(String, primary_key=True)
    auto_fill = Column(Boolean, default=True)
    initial_value = Column(String, nullable=True)
    type = Column(String, default="rasa_core.slots.UnfeaturizedSlot")

    domain = relationship("Domain", back_populates="slots")


class DomainTemplate(Base):
    """Stores the templates which are defined in the domain."""

    __tablename__ = "domain_template"

    id = Column(Integer, ForeignKey("domain.id"), primary_key=True)
    template = Column(String, primary_key=True)
    content = Column(String)

    domain = relationship("Domain", back_populates="templates")
