import logging
from typing import Text, Dict, List, Union, Optional, Set

from sqlalchemy import and_
from sqlalchemy.orm import Session

from rasax.community.database.intent import Intent, TemporaryIntentExample
from rasax.community.services.data_service import DataService
from rasax.community.services.logs_service import LogsService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.user_goal_service import (
    GOAL_INTENTS_KEY,
    GOAL_NAME_KEY)
from rasax.community.services.user_goal_service import UserGoalService

logger = logging.getLogger(__name__)

INTENT_NAME_KEY = "intent"
INTENT_MAPPED_TO_KEY = "mapped_to"
INTENT_USER_GOAL_KEY = "user_goal"
INTENT_IS_TEMPORARY_KEY = "is_temporary"
INTENT_EXAMPLES_KEY = "example_hashes"


class IntentService(object):

    def __init__(self, session: Session):
        self.session = session
        self.projects_service = ProjectsService(self.session)
        self.data_service = DataService(self.session)

    def add_temporary_intent(self,
                             intent: Dict[str, str],
                             project_id: Text) -> None:
        """Creates a new temporary intent.

        Args:
            intent: The intent object which should be added.
            project_id: The project id which the intent should belong to.
        """
        mapped_to = intent.get(INTENT_MAPPED_TO_KEY)
        permanent_intents = self.get_permanent_intents(project_id)

        if mapped_to and mapped_to not in permanent_intents:
            raise ValueError(
                "Intent cannot be mapped to '{}' since this intent"
                " does not exist.".format(mapped_to))

        intent_name = intent[INTENT_NAME_KEY]
        existing = self._get_temporary_intent(intent_name, project_id)

        if existing is not None:
            examples = existing.temporary_examples
            if INTENT_MAPPED_TO_KEY in intent.keys() and mapped_to is None:
                for e in examples:
                    self.data_service.delete_example_by_hash(project_id,
                                                             e.example_hash)
            else:
                self.data_service.update_intent(
                    mapped_to or intent_name,
                    [e.example_hash for e in examples], project_id)

        temporary = Intent(name=intent_name, mapped_to=mapped_to,
                           project_id=project_id, is_temporary=True)
        if existing:
            temporary.id = existing.id
            self.session.merge(temporary)
        else:
            self.session.add(temporary)

        self.session.commit()

    def get_temporary_intents(self, project_id: Text
                              ) -> List[Dict[str, str]]:
        """Returns all temporary intents.

        Args:
            project_id: Project id of the temporary intents.

        Returns:
            List of intent objects.
        """

        temporary_intents = self.session.query(Intent).filter(
            and_(Intent.project_id == project_id,
                 Intent.is_temporary)).all()

        return [t.as_dict() for t in temporary_intents]

    def get_temporary_intent(self, intent_id: Text, project_id: Text
                             ) -> Optional[Dict[Text, Text]]:
        """Returns a single temporary intent matching the name.

        Args:
            intent_id: Name of the temporary intent which is searched for.
            project_id: Project id which the temporary intent belongs to.

        Returns:
            The intent object if the temporary intent was found, otherwise
            `None`.
        """

        intent = self._get_temporary_intent(intent_id, project_id)

        if intent:
            return intent.as_dict()
        else:
            return None

    def _get_temporary_intent(self, intent_id: Text, project_id: Text
                              ) -> Intent:
        return self.session.query(Intent).filter(
            and_(Intent.project_id == project_id,
                 Intent.is_temporary,
                 Intent.name == intent_id)).first()

    def update_temporary_intent(self,
                                intent_id: Text,
                                intent: Dict,
                                project_id: Text):
        """Updates an existing temporary intent.

        Args:
            intent_id: Temporary intent which should be updated.
            intent: The intent object which is used to update the
                    existing intent.
            project_id: Project id of the temporary intent.
        """
        if intent.get(INTENT_NAME_KEY) != intent_id:
            raise ValueError("Intent name '{}' cannot be changed to '{}'."
                             "".format(intent_id, intent.get(INTENT_NAME_KEY)))

        self.add_temporary_intent(intent, project_id)

    def delete_temporary_intent(self,
                                intent_id: Text,
                                project_id: Text):
        """Deletes a temporary intent and related training data.

        Args:
            intent_id: Name of the temporary intent which should be deleted.
            project_id: Project id of the intent.
        """
        existing = self._get_temporary_intent(intent_id, project_id)

        if existing:
            for e in existing.temporary_examples:
                self.data_service.delete_example_by_hash(project_id,
                                                         e.example_hash)

            self.session.delete(existing)
            self.session.commit()

    def get_permanent_intents(self, project_id: Text
                              ) -> Set[Text]:
        """Returns a list of all permanent (not temporary) intents

        Args:
            project_id: Project id which these intents belong to.

        Returns:
            Set of unique permanent intents.
        """
        intents = self.projects_service.get_intents_from_domain(project_id)
        intents |= {i[INTENT_NAME_KEY] for i
                    in self.data_service.get_intents(project_id)}

        return intents

    def get_intents(self,
                    project_id: Text
                    ) -> List[Dict[str, Union[str, List[str]]]]:
        """Returns all intents including related data.

        Args:
            project_id: Project id which these intents belong to.

        Returns:
            List of intent objects including lists of `suggestions`,
            related `training_data`, and `user_goal`.
        """
        intents = self.data_service.get_intents(project_id)
        domain_intents = self.projects_service.get_intents_from_domain(
            project_id)

        flat_intents = [i[INTENT_NAME_KEY] for i in intents]

        for intent_name in domain_intents:
            if intent_name not in flat_intents:
                intents.append({INTENT_NAME_KEY: intent_name})

        logs_service = LogsService(self.session)
        for i in intents:
            suggestions, _ = logs_service.get_suggestions(
                intent_query=i[INTENT_NAME_KEY],
                fields_query=[("hash", True)])
            i["suggestions"] = [s.get('hash') for s in suggestions]

        intents += self.get_temporary_intents(project_id)

        # merge user goals
        user_goal_service = UserGoalService(self.session)
        user_goals = user_goal_service.get_user_goals(project_id)
        for g in user_goals:
            for i in intents:
                if i[INTENT_NAME_KEY] in g[GOAL_INTENTS_KEY]:
                    i[INTENT_USER_GOAL_KEY] = g[GOAL_NAME_KEY]

        return intents

    def add_example_to_temporary_intent(self,
                                        intent: Text,
                                        example_hash: Text,
                                        project_id: Text) -> None:
        """Adds a training example to a temporary intent.

        Args:
            intent: Name of the intent which the example should be added to.
            example_hash: Text hash of the training example.
            project_id: Project id of the temporary intent.
        """

        intent = self._get_temporary_intent(intent, project_id)
        if intent:
            example = TemporaryIntentExample(intent_id=intent.id,
                                             example_hash=example_hash)
            self.session.add(example)
            self.session.commit()
        else:
            logger.warning(
                "Intent '{}' was not found. Hence, the example with "
                "hash '{}' could not be deleted."
                "".format(intent, example_hash))

    def remove_example_from_temporary_intent(self,
                                             intent: Text,
                                             example_hash: Text,
                                             project_id: Text) -> None:
        """Remove a training example to a temporary intent.

        Args:
            intent: Name of the intent which the example should be removed
                from.
            example_hash: Text hash of the training example.
            project_id: Project id of the temporary intent.
        """

        temporary_intent = self._get_temporary_intent(intent, project_id)
        example = self.session.query(TemporaryIntentExample).filter(
            and_(TemporaryIntentExample.intent_id == temporary_intent.id,
                 TemporaryIntentExample.example_hash == example_hash)).first()

        self.session.delete(example)
        self.session.commit()
